/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/

#include "Constrain.h"

void buildWSC()
{
	buildWSCpa();
	buildWSCap();
	buildWSCdp();
}

void buildWSCpa() 
{	
	if(totalLeafs == NULL) return;
	for(ActionList *p = totalLeafs->head; p != NULL; p = p->next) {
		for(Action *pAction = p->head; pAction != NULL; pAction = pAction->glbNext) {
			Predicate *act = pAction->action;
			PredicateList *obs = pAction->observations;
			if(obs == NULL) continue;
		
			for(Predicate *pState = obs->head; pState != NULL; pState = pState->next) {
				if(!isPassWhiteList(pState, act)) continue;
				Integer *args = belong(pState->paras, act->paras);
				if(args == NULL) continue;
				
				Atom *atom = getActionAtom(pState, act, args, &(formula1->actionPREList), formula1->atomNum, true);
				Literal *literal = getLiteral(atom, true);
				Clause *clause = getClause(literal);
				addClauseList(&(formula1->clauseList), formula1->clauseNum, clause, 1, WSC);
			}
		}
	}
}

void buildWSCap()
{
	if(totalLeafs == NULL || goals == NULL) return;
	
	Action *guard = newAction();
	PredicateList *goal = goals->head;	
	
	for(ActionList *p1 = totalLeafs->head; p1 != NULL; p1 = p1->next) {		
		guard->observations = goal;
		p1->tail->glbNext = guard;
	
		for(Action *p = p1->head; p != guard; p = p->glbNext) {
		 	PredicateList *obs = getAddedState(p->observations, p->glbNext->observations);
	 		if(obs == NULL) continue;
	 	
		 	Predicate *act = p->action;
	 	
		 	for(Predicate *pState = obs->head; pState != NULL; pState = pState->add) {
		 		
 				Integer *args = belong(pState->paras, act->paras);
				if(args == NULL) continue;

				Atom *atom = getActionAtom(pState, act, args, &(formula1->actionADDList), formula1->atomNum);
				Literal *literal = getLiteral(atom, true);
				Clause *clause = getClause(literal);
				addClauseList(&(formula1->clauseList), formula1->clauseNum, clause, 1, WSC);
				Atom *atom1 = getActionAtom(pState, act, args, &(formula2->actionADDList), formula2->atomNum);
				Literal *literal1 = getLiteral(atom1, true);
				Clause *clause1 = getClause(literal1);
				addClauseList(&(formula2->clauseList), formula2->clauseNum, clause1, 1, WSC);
			}
			free(obs); 
		 } 
		p1->tail->glbNext = NULL;
		goal = goal->next;
	}
	free(guard);
}

void buildWSCdp()
{
	if(totalLeafs == NULL || goals == NULL) return;
	
	Action *guard = newAction();
	PredicateList *goal = goals->head;
	
	for(ActionList *p1 = totalLeafs->head; p1 != NULL; p1 = p1->next) {
		guard->observations = goal;
		p1->tail->glbNext = guard;
	
		for(Action *p = p1->head; p != guard; p = p->glbNext) {
	 		PredicateList *obs = getAddedState(p->glbNext->observations, p->observations);
	 		if(obs == NULL) continue;
	 	
		 	Predicate *act = p->action;
		 	for(Predicate *pState = obs->head; pState != NULL; pState = pState->add) {
		 		if(isInBlackList(pState, act)) continue;
				Integer *args = belong(pState->paras, act->paras);
				if(args == NULL) continue;
				
				Atom *atom = getActionAtom(pState, act, args, &(formula1->actionDELList), formula1->atomNum);
				Literal *literal = getLiteral(atom, true);
				Clause *clause = getClause(literal);
				addClauseList(&(formula1->clauseList), formula1->clauseNum, clause, 1, WSC);
				Atom *atom1 = getActionAtom(pState, act, args, &(formula2->actionDELList), formula2->atomNum);
				Literal *literal1 = getLiteral(atom1, true);
				Clause *clause1 = getClause(literal1);
				addClauseList(&(formula2->clauseList), formula2->clauseNum, clause1, 1, WSC);
			}
			free(obs);	
		 } 
		p1->tail->glbNext = NULL;
		goal = goal->next;
	}
	free(guard);
}

void buildWPC()
{
	if(totalLeafs == NULL) return;
		
	for(ActionList *p = totalLeafs->head; p != NULL; p = p->next) {
		buildWPCFromAction(p);	
	}
	
}

void buildWPCFromAction(ActionList *actionList)
{
	if(actionList == NULL) return;
	for(Action *p = actionList->tail; p != actionList->head; p = p->glbPre) {
		PredicateList *list = getVirtualStates(p->action->paras);
		if(list == NULL) continue;
		
	   for(Predicate *p1 = list->head; p1 != NULL; p1 = p1->next) {
			Integer *args = belong(p1->paras, p->action->paras);
			if(args == NULL) continue;
			if(!isPassWhiteList(p1->name, args, p->action)) continue;
			LiteralTable *table = newLiteralTable();
			
			Atom *atom = getActionAtom(p1, p->action, args, &(formula1->actionPREList), formula1->atomNum);
			table->head = newLiteralList();
			table->tail = table->head;
			addLiteral(table->tail, getLiteral(atom, false));
			
		   for(Action *pAction = p->glbPre; pAction != NULL; pAction = pAction->glbPre) {
				Integer *args1 = belong(p1->paras, pAction->action->paras);
				if(args1 == NULL) continue;
				
				Atom *atom1 = getActionAtom(p1, pAction->action, args1, &(formula1->actionADDList), formula1->atomNum);
				table->tail->next = newLiteralList();
				table->tail = table->tail->next;
				addLiteral(table->tail, getLiteral(atom1, true));
				
				Action *pAct = pAction->glbNext;
				while(pAct != p) {
					Integer *paras = belong(p1->paras, pAct->action->paras);
					if(paras != NULL) {
						if(!isInBlackList(p1->name, paras, pAct->action)) {
							Atom *atom2 = getActionAtom(p1, pAct->action, paras, &(formula1->actionDELList), formula1->atomNum);
							addLiteral(table->tail, getLiteral(atom2, false));
						}
					}
					pAct = pAct->glbNext;
				} 
			} 
			
			bool found =false;
			for(Predicate *pFirst = actionList->head->observations->head; pFirst != NULL; pFirst = pFirst->next) {
				if(equal(p1, pFirst)) {
					found = true;
					break;
				}
			}
			if(found) {
				LiteralList *list = NULL;
				for(Action *pAction = actionList->head; pAction != NULL; pAction = pAction->glbNext) {
					Integer *paras = belong(p1->paras, pAction->action->paras);
					if(paras != NULL) {
						if(!isInBlackList(p1->name, paras, pAction->action)) {
						Atom *atom = getActionAtom(p1, pAction->action, paras, &(formula1->actionDELList), formula1->atomNum);
							Literal *literal = getLiteral(atom, false);
							if(list == NULL) {
								list = newLiteralList();
							}
							addLiteral(list, literal);
						}
					}
				}
				
				if(list != NULL) {
					table->tail->next = newLiteralList();
					table->tail = table->tail->next;
					table->tail->head = list->head;
					table->tail->tail = list->tail;
				}
			}
			if(table->head == table->tail) continue;
			LiteralTable *tableCNF = getCNFClauseList(table);			
			if(tableCNF == NULL) continue;
			for(LiteralList *pCNF = tableCNF->head; pCNF != NULL; pCNF = pCNF->next) {
				Clause *clause = newClause();
				clause->literalList = pCNF;
				addClauseList(&(formula1->clauseList), formula1->clauseNum, clause, formula1->maxWeight);
			} 
		}
	}
}

void buildWAC(Formula *formula)
{
	buildWACAP(formula);
	buildWACDP(formula);
}

void buildWACAP(Formula *formula)
{
	if(formula->actionADDList == NULL || formula->actionPREList == NULL) return;
	
	for(Atom *p1 = formula->actionADDList->head; p1 != NULL; p1 = p1->next) {
		bool found = false;
		
		Atom *p2 = formula->actionPREList->head;
		for(; p2 != NULL; p2 = p2->next) {
			if(p1->ptr1 != p2->ptr1 || p1->ptr2 != p2->ptr2) continue;
			if(!equal(p1->args, p2->args)) continue;
			found = true;
			break;
		}
		if(!found) continue;
		
		Literal *literal1 = getLiteral(p1, false);
		Literal *literal2 = getLiteral(p2, false);
		Clause *clause = getClause(literal1, literal2);
		addClauseList(&(formula->clauseList), formula->clauseNum, clause, formula->maxWeight);
	}
}

void buildWACDP(Formula *formula)
{	
	if(formula->actionDELList == NULL || formula->actionPREList == NULL) return;
	
	for(Atom *p1 = formula->actionDELList->head; p1 != NULL; p1 = p1->next) {
		bool found = false;
		
		Atom *p2 = formula->actionPREList->head;
		for(; p2 != NULL; p2 = p2->next) {
			if(p1->ptr1 != p2->ptr1 || p1->ptr2 != p2->ptr2) continue;
			if(!equal(p1->args, p2->args)) continue;
			found = true;
			break;
		}
		if(!found) continue;
		
		Literal *literal1 = getLiteral(p1, false);
		Literal *literal2 = getLiteral(p2, true);
		Clause *clause = getClause(literal1, literal2);
		addClauseList(&(formula->clauseList), formula->clauseNum, clause, formula->maxWeight);
	}
}






