/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/



#ifndef _DATASTRUCT_
#define _DATASTRUCT_

#define MAX_LENGTH 256
#define MAX_VAR 128
#define MAX_VAR_LEN 64

typedef struct _Variable {
	char *var;
	struct _Variable *next;
} Variable;

typedef struct _VarArray{
	char vars[MAX_VAR][MAX_VAR_LEN];
	int len;
} VarArray;

typedef struct _Integer {
	int *integers;
	int len;
	struct _Integer *next;
} Integer;

typedef struct _VariableList {
	Variable *head;
	Variable *tail;
	struct _VariableList *next;
} VariableList;

typedef struct _VariableTable {
	VariableList *head;
	VariableList *tail;
} VariableTable;

typedef struct _Object {
	char* name;
	char* type;
	struct _Object *next;
} Object;

typedef struct _ObjectList {
	Object *head;
	Object *tail;
} ObjectList;

typedef struct _Predicate {
	char *name;
	VariableList *paras;
	struct _Predicate *next;
	struct _Predicate *add;
} Predicate;

typedef struct _PredicateList {
	Predicate *head;
	Predicate *tail;
	struct _PredicateList *next;
} PredicateList;

typedef struct _PredicateTable {
	PredicateList *head;
	PredicateList *tail;
} PredicateTable;

typedef struct _TaskStruct {
	char *name;
	int vars[MAX_VAR];
	struct _TaskStruct *next;
} TaskStruct;

typedef struct _TaskStructList {
	TaskStruct *head;
	TaskStruct *tail;
} TaskStructList;  

typedef struct _TaskList TaskList;
typedef struct _LeafPointer LeafPointer;

typedef struct _MethodStruct {
	Predicate *method;
	TaskStruct *task;
	TaskStructList *subTasks;
	TaskList *taskList;
	struct _MethodStruct *next;
} MethodStruct;

typedef struct _MethodStructList {
	MethodStruct *head;
	MethodStruct *tail;
} MethodStructList;

typedef struct _ActionList ActionList;

typedef struct _Task {
	int order;
	Predicate *task;
	struct _Task *preTask;
	ActionList *actions;
	TaskList *subTasks;
	VariableList *methodParas;
	MethodStruct *methodStruct;
	struct _Task *next;
	LeafPointer *leafPointer;
	int tried;
	struct _Task *link;
} Task;

struct _TaskList {
	Task *head;
	Task *tail;
};

typedef struct _ActionModel ActionModel;

typedef struct _Action {
	int order;	
	Predicate *action;
	PredicateList *observations;
	Task *preTask;
	ActionModel *actionModel;
	struct _Action *next;
	struct _Action *glbNext;
	struct _Action *pre;
	struct _Action *glbPre;
} Action;

struct _ActionList {
	Action *head;
	Action *tail;
	struct _ActionList *next;
};

typedef struct _ActionTable {
	ActionList *head;
	ActionList *tail;
} ActionTable;

typedef struct _Wrap {
	VariableList *list;
	Task *task;
	MethodStruct *mStruct;
} Wrap;

typedef struct _Node {
	VariableList *varList;
	Action *action;
	Task *task;
	struct _Node *next;
} Node;

typedef struct _NodeList {
	Node *head;
	Node *tail;
} NodeList;

typedef struct _Atom {
	bool obs;
	int no;
	void *ptr1;
	void *ptr2;
	Integer* args;
	struct _Atom *next;
} Atom;

typedef struct _AtomList {
	Atom *head;
	Atom *tail;
} AtomList;

typedef struct _Literal {
	bool isPositive;
	Atom *atom;
	struct _Literal *next;
} Literal;

typedef struct _LiteralList {
	Literal *head;
	Literal *tail;
	struct _LiteralList *next;
} LiteralList;

typedef struct _LiteralTable {
	LiteralList *head;
	LiteralList *tail;
} LiteralTable;

typedef struct _Clause {
	int weight;
	int finalWeight;
	LiteralList *literalList;
	struct _Clause *next;
} Clause;

typedef struct _ClauseList {
	Clause *head;
	Clause *tail;
} ClauseList;

typedef struct _MethodName {
	char *name;
	int count;
	struct _MethodName *next;
} MethodName;

typedef struct _MethodNameList {
	MethodName *head;
	MethodName *tail;
} MethodNameList;

typedef struct _FormalState {
	char *name;
	Integer *paras;
	struct _FormalState *next;
} FormalState;

typedef struct _FormalStateList {
	FormalState *head;
	FormalState *tail;
} FormalStateList;

typedef struct _ActionPred {
	char *actionName;
	FormalStateList *stateList;
	struct _ActionPred *next;
} ActionPred;

typedef struct _ActionPredList {
	ActionPred *head;
	ActionPred *tail;
} ActionPredList;

struct _LeafPointer {
	Action *start;
	Action *end;
};

struct _ActionModel {
	Predicate *action;
	FormalStateList *precondList;
	FormalStateList *addList;
	FormalStateList *delList;
	struct _ActionModel *next;
};

typedef struct _ActionModelList {
	ActionModel *head;
	ActionModel *tail;
} ActionModelList;

typedef struct _MethodModel {
	MethodStruct *methodStruct;
	FormalStateList *precondList;
	FormalStateList *blackList;
	struct _MethodModel *next;
} MethodModel;

typedef struct _MethodModelList {
	MethodModel *head;
	MethodModel *tail;
} MethodModelList;

typedef struct _Formula {
	AtomList *actionPREList;
	AtomList *actionADDList;
	AtomList *actionDELList;
	ClauseList *clauseList;
	int atomNum;
	int maxWeight;
	int clauseNum;
} Formula;

typedef struct _tMethod {
	char *name;
	ObjectList *paras;
	Predicate *task;
	PredicateList *subTasks;
	PredicateList *precond;
	struct _tMethod *next;
} tMethod;

typedef struct _tMethodList {
	tMethod *head;
	tMethod *tail;
} tMethodList;

typedef struct _tAction {
	char *name;
	ObjectList *paras;
	PredicateList *precond;
	PredicateList *addList;
	PredicateList *delList;
	struct _tAction *next;
} tAction;

typedef struct _tActionList {
	tAction *head;
	tAction *tail;
} tActionList;

typedef struct _tVariableList {
	bool flag;
	VariableList *variableList;
	struct _tVariableList *next;
} tVariableList;

typedef struct _tVariableTable {
	tVariableList *head;
	tVariableList *tail;
} tVariableTable;

#endif
