/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/


#include "IOMan.h"

void loadInputFile(FILE **file, char *filename)
{
	FILE *f;	
	f = fopen(filename, "r");
	if(!f) { 
		printf("Can't open file: %s!\n", filename); 
		return; 
	}
	
	*file = f;
	Solutionparse();
	
	fclose(f);
}

void readProfile(char path[], int &num, char maxsatName[], char maxsatName1[], char resName[])
{
	FILE *file = fopen("Profile", "r");
	if(!file) return;
	fscanf(file, "%s", path);
	fscanf(file, "%d", &num);
	fscanf(file, "%f", &partObsPercent);
	fscanf(file, "%f", &partDtrPercent);
	fscanf(file, "%s", maxsatName);
	fscanf(file, "%s", maxsatName1);
	fscanf(file, "%s", resName);
}

void outputMaxsat(Formula *formula, char *filename)
{
	if(formula->clauseList == NULL) return;
	
	FILE *file = fopen(filename, "w");
	fprintf(file, "p wcnf %d %d\n", formula->atomNum, formula->clauseNum);
	
	for(Clause* p1 = formula->clauseList->head; p1 != NULL; p1 = p1->next) {
		fprintf(file, "%d    ", p1->finalWeight);
		for(Literal *p2 = p1->literalList->head; p2 != NULL; p2 = p2->next) {
			if(p2->isPositive) fprintf(file, "%d   ", p2->atom->no);
			else fprintf(file, "-%d   ", p2->atom->no);
		}
		fprintf(file, "0\n");
	}
	
	fclose(file);	
}

void outputResult(char *filename)
{
	FILE *file = fopen(filename, "w");
	fprintf(file, "(define\n(:requirements :typing)\n(:types");
	outputType(file);
	fprintf(file, ")\n(:predicates\n");
	outputPredicate(file);
	fprintf(file, ")\n\n");
	outputMethodList(file);
	outputActionList(file);
	fprintf(file, ")");
	fclose(file);
}

void outputType(FILE *file)
{
	VariableList *typeList = NULL;
	for(Object *p = objectList->head; p != NULL; p = p->next) {
		if(typeList == NULL) {
			typeList = newVariableList();
			typeList->head = newVariable();
			typeList->tail = typeList->head;
			typeList->tail->var = newStr(strlen(p->type)+1);	
			strcpy(typeList->tail->var, p->type);
			continue;
		}
		bool found = false;
		for(Variable *p1 = typeList->head; p1 != NULL; p1 = p1->next) {
			if(strcmp(p->type, p1->var) == 0) {
				found = true;
				break;
			}
		}
		if(found) continue;
		typeList->tail->next = newVariable();
		typeList->tail = typeList->tail->next;
		typeList->tail->var = newStr(strlen(p->type)+1);
		strcpy(typeList->tail->var, p->type);
	}
	for(Variable *p = typeList->head; p != NULL; p = p->next) {
		fprintf(file, " %s", p->var);
	}
}

void outputPredicate(FILE *file)
{
	if(formalPredList == NULL) return;
	for(Predicate *p = formalPredList->head; p != NULL; p = p->next) {
		fprintf(file, "(");
		fprintf(file, "%s", p->name);
		if(p->paras != NULL) {
			char ch = 'a';
			for(Variable *p1 = p->paras->head; p1 != NULL; p1 = p1->next) {
				fprintf(file, " ?%c -%s", ch++, p1->var);
			}
		}
		fprintf(file, ")\n");
	}
}

void outputMethodList(FILE *file)
{
	if(methodModelList == NULL) return;
	for(MethodModel *p = methodModelList->head; p != NULL; p = p->next) {
		printMethodName(file, p->methodStruct->method->name);
		fprintf(file, ":parameters");
		if(p->methodStruct->method->paras != NULL) {
			fprintf(file, "(");
			char ch = 'a';
			for(Variable *p1 = p->methodStruct->method->paras->head; p1 != NULL; p1 = p1->next) {
				fprintf(file, "?%c -%s ", ch++, p1->var);
			}
			fprintf(file, ")");
		}
		fprintf(file, "\n:task (%s", p->methodStruct->task->name);
		int i = 0;
		while(p->methodStruct->task->vars[i]) {
			fprintf(file, " ?%c", 'a'+p->methodStruct->task->vars[i]-1);
			++i;
		}
		
		fprintf(file, ")\n:precondition (and ");
		printFormalStateList(file, p->precondList, true);		
		fprintf(file, ")\n");
		
		if(p->methodStruct->subTasks != NULL) {
			fprintf(file, ":subtasks (and ");
			for(TaskStruct *p1 = p->methodStruct->subTasks->head; p1 != NULL; p1 = p1->next) {
				fprintf(file, "(%s", p1->name);
				int i = 0;
				while(p1->vars[i]) {
					fprintf(file, " ?%c", 'a'+p1->vars[i]-1);
					++i;
				}
				if(p1 != p->methodStruct->subTasks->tail) fprintf(file, ") ");
				else fprintf(file, ")");
			}
			fprintf(file, ")\n");
		}
		fprintf(file, ")\n\n");
	}
}

void printMethodName(FILE *file, char *name)
{
	if(methodNameList == NULL) {
		methodNameList = newMethodNameList();
		methodNameList->head = newMethodName();
		methodNameList->tail = methodNameList->head;
		copystr(&(methodNameList->head->name), name);
		++methodNameList->head->count;
		fprintf(file, "(:method %s-m%d\n", name, methodNameList->head->count);
		return;
	}
	bool found = false;
	int count;
	for(MethodName *p = methodNameList->head; p != NULL; p = p->next) {
		if(strcmp(name, p->name) == 0) {
			++p->count;
			found = true;
			count = p->count;
			break;
		}
	}
	if(found) {
		fprintf(file, "(:method %s-m%d\n", name, count);
		return;
	}
	methodNameList->tail->next = newMethodName();
	methodNameList->tail = methodNameList->tail->next;
	copystr(&(methodNameList->tail->name), name);
	++methodNameList->tail->count;
	fprintf(file, "(:method %s-m%d\n", name, 1);
}

void outputActionList(FILE *file)
{
	if(actionModelList == NULL) return;
	for(ActionModel *p = actionModelList->head; p != NULL; p = p->next) {
		fprintf(file, "(:action %s\n", p->action->name);	
		fprintf(file, ":parameters ");
		if(p->action->paras != NULL) {
			fprintf(file, "(");
			char ch = 'a';
			for(Variable *p1 = p->action->paras->head; p1 != NULL; p1 = p1->next) {
				fprintf(file, "?%c -%s ", ch++, p1->var);
			}
			fprintf(file, ")");
		}
		fprintf(file, "\n");
		
		fprintf(file, ":precondition (and ");
		printFormalStateList(file, p->precondList, true);
		fprintf(file, ")\n");
		
		fprintf(file, ":effect (and ");
		printFormalStateList(file, p->delList, false);
		printFormalStateList(file, p->addList, true);
		fprintf(file, ")\n)\n\n");
	}
}

void printFormalStateList(FILE *file, FormalStateList *list, bool isPositive) 
{
	if(list == NULL) return;
	
	for(FormalState *p = list->head; p != NULL; p = p->next) {
		if(!isPositive) fprintf(file, "(not");
		fprintf(file, "(%s", p->name);
		if(p->paras != NULL && p->paras->len != 0) {
			fprintf(file, " ");
			for(int i = 0; i != p->paras->len-1; ++i) {
				fprintf(file, "?%c ", *(p->paras->integers+i)+'a'-1);
			}
			fprintf(file, "?%c", *(p->paras->integers+p->paras->len-1)+'a'-1);
		}
		if(!isPositive) fprintf(file, ")) ");
		else {
			if(p != list->tail) fprintf(file, ") ");
			else fprintf(file, ")");
		}
	}
}
