/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/


#ifndef _MAXSATZ_
#define _MAXSATZ_

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <sys/times.h>
#include <sys/types.h>
#include <limits.h>

#include <unistd.h>

typedef signed char my_type;
typedef unsigned char my_unsigned_type;
typedef long long int lli_type;

#define WORD_LENGTH 1024
#define TRUE 1
#define FALSE 0
#define NONE -1
#define WEIGHT 4
#define WEIGHT1 25
#define WEIGHT2 5
#define WEIGHT3 1
#define T 10

/* the tables of variables and clauses are statically allocated. Modify the 
   parameters tab_variable_size and tab_clause_size before compilation if 
   necessary */
#define tab_variable_size  30000
#define tab_clause_size 1000000
#define tab_unitclause_size ((tab_clause_size/4<2000) ? 2000 : tab_clause_size/4)
#define my_tab_variable_size ((tab_variable_size/2<1000) ? 1000 : tab_variable_size/2)
#define my_tab_clause_size ((tab_clause_size/2<2000) ? 2000 : tab_clause_size/2)
#define my_tab_unitclause_size ((tab_unitclause_size/2<1000) ? 1000 : tab_unitclause_size/2)
#define tab_literal_size 2*tab_variable_size
#define double_tab_clause_size 2*tab_clause_size
#define positive(literal) literal<NB_VAR
#define negative(literal) literal>=NB_VAR
#define get_var_from_lit(literal) \
  ((literal<NB_VAR) ? literal : literal-NB_VAR)
#define complement(lit1, lit2) \
 ((lit1<lit2) ? lit2-lit1 == NB_VAR : lit1-lit2 == NB_VAR)
#define get_lit(v, s) ((s == POSITIVE) ? v : v + NB_VAR)
#define tab_clause_size_XL (tab_clause_size * 10)
#define inverse_signe(signe) \
 (signe == POSITIVE) ? NEGATIVE : POSITIVE
#define unsat(val) (val==0)?"UNS":"SAT"
#define pop(stack) stack[--stack ## _fill_pointer]
#define push(item, stack) stack[stack ## _fill_pointer++] = item
#define top(stack) stack[stack ## _fill_pointer - 1]
#define min(a, b)  (((a) < (b)) ? (a) : (b))
#define NEGATIVE 0
#define POSITIVE 1
#define PASSIVE 0
#define ACTIVE 1
#define INIT_BASE_NB_CLAUSE (tab_clause_size / 2)
#define NO_CONFLICT -3
#define NO_REASON -3
#define max_var_learned (tab_variable_size / 10)
#define MAX_LEN_LEARNED 20 // MAX_LEN_LEARNED = num_lits * 2 (best performance 20)
#define OLD_CLAUSE_REDUNDANT -77
#define NEW_CLAUSE_REDUNDANT -7
#define SIMPLE_NON_LINEAR_CASE 2
#define SIMPLE_RS1_NON_LINEAR_CASE 3
#define SIMPLE_RS2_NON_LINEAR_CASE 4
#define SIMPLE_CUB_NON_LINEAR_CASE 5
#define SIMPLE_RS1_3_NON_LINEAR_CASE 6
#define SIMPLE_RS2_3_NON_LINEAR_CASE 7
#define OTHER_LIT_FIXED 1
#define THIS_LIT_FIXED 2
#define CLK_TCK 1000   //Modified

int treat_complementary_unitclauses(int var, int clause);
void add_new_lit_in(int *pos, int clause);
void add_new_hlit_in(int *pos, int clause);
int get_next_clause(int *pos);
int get_prev_clause(int *pos);
int smaller_than(int lit1, int lit2);
int redundant(int *new_clause, int *old_clause);
void remove_passive_clauses();
void remove_passive_vars_in_clause(int clause);
int clean_structure();
void lire_clauses(FILE *fp_in, int instance_type);
void build_structure();
void eliminate_redundance();
int build_simple_sat_instance(char *input_file);
void print_clause(int clause);
void remove_clauses(int var);
int reduce_clauses(int var);
int my_reduce_clauses(int var);
int my_reduce_clauses_for_fl(int var);
void print_values(int nb_var);
int backtracking();
int verify_solution();
void reset_context(int saved_clause_stack_fill_pointer, 
		   int saved_reducedclause_stack_fill_pointer, 
		   int saved_unitclause_stack_fill_pointer, 
		   int saved_variable_stack_fill_pointer);
void create_binaryclause(int var1, int sign1, 
			 int var2, int sign2, 
			 lli_type min_weight);
int verify_binary_clauses(int *varssigns, int var1, int sign1, 
			  int var2, int sign2);
int search_linear_reason1(int var);
int search_linear_reason2(int var);
int check_reason(int *varssigns, int clause, int clause1, int clause2);
int create_complementary_binclause(int clause, int clause1, 
				   int clause2, lli_type min_weight);
int get_satisfied_literal(int clause);
int create_new_unitclause(int var, int sign, lli_type min_weight);
void create_ternary_clauses(int var1, int sign1, 
			    int var2, int sign2, 
			    int var3, int sign3, 
			    lli_type min_weight);
int non_linear_conflict(int empty_clause, 
			int var1, int sign1, 
			int var2, int sign2, 
			lli_type min_weight);
int non_linear_conflict_rs1(int empty_clause, 
			    int var1, int sign1, 
			    int var2, int sign2, 
			    lli_type min_weight);
int non_linear_conflict_rs2(int empty_clause, 
			    int var1, int sign1, 
			    int var2, int sign2, 
			    lli_type min_weight);
int non_linear_conflict_cub(int empty_clause, 
			    int var1, int sign1, 
			    int var2, int sign2, 
			    lli_type min_weight);
int non_linear_conflict_rs1_3(int empty_clause, 
			      int var1, int sign1, 
			      int var2, int sign2, lli_type min_weight);
int non_linear_conflict_rs2_3(int empty_clause, int var1, int sign1, int var2, int sign2, lli_type min_weight);
int compute_minweight_for_linear_reasons();
int linear_conflict(int clause,  lli_type *min_weight);
void reduce_clause_weight(int clause, int weight);
void remove_linear_reason(int clause, lli_type min_weight);
void remove_linear_reasons(lli_type min_weight);
void remove_clauses_for_fl(int var);
int my_unitclause_process(int);
int assign_and_unitclause_process( int var, int value, int starting_point);
int store_reason_clauses( int clause);
void remove_reason_clauses(lli_type *min_weight);
lli_type simple_get_pos_clause_nb(int);
lli_type simple_get_neg_clause_nb(int);
int search_linear_reason1_for_fl(int falsified_var, int testing_var);
int search_linear_reason2_for_fl(int falsified_var, int testing_var);
void cycle_resolution(int var1, int sign1, int var2, int sign2, int var3, int sign3,
		      int clause1, int clause2, int clause3);
int simple_cycle_case(int testing_var);
int cycle_conflict(int clause, int testing_var);
void remove_replaced_clauses();
void mark_conflict_clauses();
void unmark_conflict_clauses();
int test_value(int var, int value, int saved_unitclause_stack_fill_pointer);
int lookahead_by_fl( lli_type nb_known_conflicts );
int search_linear_reason1_for_up(int falsified_var);
int search_linear_reason2_for_up(int falsified_var);
void mark_variables(int saved_variable_stack_fill_pointer);
int cycle_conflict_by_up(int clause);
int lookahead_by_up(lli_type nb_known_conflicts);
void remove_premiss_clauses_weight();
void reset_clause_weight(int saved_clause_weight_stack_fill_pointer);
int lookahead();
int satisfy_unitclause(int unitclause);
int my_unitclause_process(int starting_point);
int get_complement(int lit);
void create_unitclause(int lit, int subsumedclause, 
		       int p_clause, lli_type weight);
void verify_resolvent(int lit, int clause1, int clause2);
lli_type simple_get_pos_clause_nb(int var);
void insert_clause(int lit, int clause);
void merge_clause(int clause1, int clause2);
void mark_literals(int var);
void treat_binary_clauses(int lit, int clause1, int clause2);
int apply_rule1(int var);
int apply_rule2(int var);
int rules1_and_2();
int get_neg_clause_nb(int var);
int treat_complementary_unitclauses(int var, int clause);
int get_pos_clause_nb(int var);
int satisfy_literal(int lit);
int assign_value(int var, int current_value, int rest_value);
int add_to_learned_clause(int clause, int unit_var);
void add_clause_to_DB();
void hard_learning();
int unitclause_process();
void search_unit_from_binary();
int choose_and_instantiate_variable();
void dpl();
void init();
void ubcsat(char file[]);
int maxSatz(char *saved_input_file);

#endif
