/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/



#ifndef _MEMMAN_
#define _MEMMAN_

#include <stdlib.h>
#include <stdio.h>

#include "Datastruct.h"

/* Check allocated memory
 */
#define CHECK_MEM(p) if(NULL == (p)){ fprintf(stdout, "\n\aNO MEMORY in file %s:%d\n\n", __FILE__, __LINE__); exit(1);}
 
int* newInt(int len);
char* newStr(int len);

Variable* newVariable();
VariableList* newVariableList();
VariableTable* newVariableTable();

VarArray* newVarArray();
Integer* newInteger();

Object* newObject();
ObjectList* newObjectList();

Predicate* newPredicate();
PredicateList *newPredicateList();
PredicateTable *newPredicateTable();

TaskStruct* newTaskStruct();
TaskStructList* newTaskStructList();

MethodStruct* newMethodStruct();
MethodStructList* newMethodStructList();

Task* newTask();
TaskList* newTaskList();

Action* newAction();
ActionList* newActionList();
ActionTable* newActionTable();

Wrap* newWrap();

Node* newNode();
NodeList *newNodeList();

Atom* newAtom();
AtomList* newAtomList();

Literal* newLiteral();
LiteralList* newLiteralList();
LiteralTable* newLiteralTable();

Clause* newClause();
ClauseList* newClauseList();

MethodName* newMethodName();
MethodNameList* newMethodNameList();

FormalState* newFormalState();
FormalStateList* newFormalStateList();

ActionPred* newActionPred();
ActionPredList* newActionPredList();

LeafPointer* newLeafPointer();

ActionModel* newActionModel();
ActionModelList* newActionModelList();

MethodModel* newMethodModel();
MethodModelList* newMethodModelList();

Formula* newFormula();

tMethod* newtMethod();
tMethodList* newtMethodList();

tAction* newtAction();
tActionList* newtActionList();

tVariableList* newtVariableList();
tVariableTable* newtVariableTable();

void freePredicate(Predicate *pred);
void freeVariableList(VariableList *list);
void freeVariable(Variable *var);
void freeTaskStruct(TaskStruct *tStruct);
void freeTaskStructList(TaskStructList *list);
void freeMethodStruct(MethodStruct *mStruct);
void freeInteger(Integer *integer);
void freeAtom(Atom *atom);
void freeClause(Clause *clause);
void freeLiteralList(LiteralList *list);
void freeLiteral(Literal *literal);
void freeObjectList(ObjectList *list);
void freeObject(Object *obj);

#endif
