%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Transformation.h"
#include "MemMan.h"

extern int yylex();
void Solutionerror(char *s);

%}

%union {
	#include "Datastruct.h"
	#include "MemMan.h"
	
	char str[MAX_LENGTH];
	Predicate *pPredicate;
	VariableList *pVariableList;
	VariableTable *pVariableTable;
	Wrap *pWrap;
	Node *pNode;
	NodeList *pNodeList;
}

%token OPEN
%token CLOSE
%token OPEN_PAREN
%token CLOSE_PAREN
%token SOLN_TOK
%token OBJECT_TOK
%token INIT_TOK
%token OBSERVATION_TOK
%token GOAL_TOK

%token <str> VARIABLE
%token <str> TYPE

%type <pVariableList> variableList
%type <pVariableList> statement
%type <pVariableTable> observationList
%type <pVariableTable> statementList

%type <pNode> node
%type <pNodeList> nodeList
%type <pWrap> dtr

%%

solutionDefinition:	OPEN_PAREN SOLN_TOK objectDefinition initDefinition dtrSet goalDefinition CLOSE_PAREN
;

objectDefinition:	OPEN_PAREN OBJECT_TOK objectTable CLOSE_PAREN
;

objectTable:	objectList
|	objectTable objectList
;

objectList:	variableList TYPE
{	
	buildObjectList($1, $2);
}
;

variableList:
{
	$$ = NULL;
}
|	VARIABLE	
{
	$$ = newVariableList();
	buildVariableList($$, $1);
}
|	variableList VARIABLE	
{
	buildVariableList($1, $2);
}
;	

initDefinition:	OPEN_PAREN INIT_TOK statementList CLOSE_PAREN
;

statementList:	statement
{
	buildFormalPredList($1);
	buildActualPredList($1);
	$$ = newVariableTable();
	buildVariableTable($$, $1);
}
|	statementList statement
{
	buildFormalPredList($2);
	buildActualPredList($2);
	buildVariableTable($1, $2);
}
;

statement:	OPEN_PAREN variableList CLOSE_PAREN
{
	$$ = $2;
}
;

dtrSet:	dtr
{
	buildDtrList($1->task);
}
|	dtrSet dtr
{
	buildDtrList($2->task);
}
|	observationList statement
{
	buildActionSchemaList($2);
	addTopLeaf($1, $2);
}
|	dtrSet observationList statement
{
	buildActionSchemaList($3);
	addTopLeaf($2, $3);
}
;

dtr:	statement OPEN nodeList CLOSE
{
	$$ = newWrap();
	$$->list = $1;
	
	if(isComplete()) {
		$$->mStruct = addMethodStruct($1, getVariableTable($3));
		$$->task = getPartTask($3);
		$$->task->task = getActualPredicate($1);
		$$->task->methodParas = getMethodParas($1, $3);
		$$->task->methodStruct = $$->mStruct;
		buildMethodStruct($$->task, $$->mStruct);
		
	} else {
		$$->task = newTask();
		$$->task->task = getActualPredicate($1);
		addTopLeaf($3);
		addTopTask($3);
	}
}
|	statement OPEN CLOSE
{
	$$ = newWrap();
	$$->list  = $1;
	$$->task = newTask();
	$$->task->task = getActualPredicate($1);
}
;

nodeList:	node
{
	$$ = newNodeList();
	buildNodeList($$, $1);
}
|	nodeList node
{
	buildNodeList($1, $2);
}
;

node:	observationList statement 
{
	buildActionSchemaList($2);
	
	$$  = newNode();
	$$->varList = $2; 
	$$->action = getPartAction($1, $2);
	addLeaf($$->action);
}
|	dtr 
{
	$$ = newNode();
	$$->varList = $1->list;
	$$->task = $1->task;
}
;

observationList:	OPEN_PAREN OBSERVATION_TOK statementList CLOSE_PAREN
{
	$$ = getPartObservations($3);
}
;

goalDefinition:	OPEN_PAREN GOAL_TOK statementList CLOSE_PAREN
{
	buildGoal($3);
}
;

%%

void Solutionerror(char *s)
{
	fprintf(stderr, "%s\n", s);
}
