/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/


#include "Tool.h"

void copystr(char **des, char *src)
{
	*des = newStr(strlen(src)+1);
	strcpy(*des, src);
}

void buildMethodParas(VariableList *ans, VariableList *paras)
{
	if(paras == NULL) return;
	
	for(Variable *p = paras->head; p != NULL; p = p->next) {
		bool repeat = false;
		for(Variable *p1 = ans->head; p1 != NULL; p1 = p1->next) {
			if(strcmp(p->var, p1->var) == 0) {
				repeat = true;
				break;
			}
		}
		if(repeat) continue;
		
		Variable *pTail = ans->tail;
		pTail->next = newVariable();
		pTail = pTail->next;
		copystr(&(pTail->var), p->var);
	}
}

void buildTaskStruct(TaskStruct *task, VariableList *list, VarArray &paras)
{
	copystr(&(task->name), list->head->var);
	
	int i = 0;
	for(Variable *pVar = list->head->next; pVar != NULL; pVar = pVar->next) {
		int j = 0;
		
		bool flag = false;
		for(; j < paras.len; ++j) {
			if(strcmp(pVar->var, paras.vars[j]) == 0) {
				task->vars[i] = j+1;
				++i;
				flag = true;
				break;
			}
		}
		if(flag) continue;
		
		strcpy(paras.vars[j], pVar->var);
		task->vars[i] = j+1;
		++i;
		++paras.len;
	}
}

void buildTaskOfMethod(MethodStruct *mStruct, VariableList *list, VarArray &paras) 
{
	mStruct->task = newTaskStruct();
	buildTaskStruct(mStruct->task, list, paras);
}

void buildSubTasksOfMethod(MethodStruct *mStruct, VariableTable *table, VarArray &paras)
{
	mStruct->subTasks = newTaskStructList();
	mStruct->subTasks->head = newTaskStruct();
	mStruct->subTasks->tail = mStruct->subTasks->head;
	
	for(VariableList *pList = table->head; pList != NULL; pList = pList->next) {
		buildTaskStruct(mStruct->subTasks->tail, pList, paras);
		if(pList->next != NULL) {
			mStruct->subTasks->tail->next = newTaskStruct();
			mStruct->subTasks->tail = mStruct->subTasks->tail->next;
		}
	}
}

Predicate* findPredicate(Predicate *state)
{
	return findPred(state, formalPredList);
}

Predicate* findActionSchema(Predicate *action)
{
	return findPred(action, actionSchemaList);
}

Predicate* findPred(Predicate *pred, PredicateList *list)
{
	if(list == NULL) return NULL;
	Predicate *tmp = getFormalPredicate(pred);
	Predicate *ptr = list->head;
	bool found = false;
	for(; ptr != NULL; ptr = ptr->next) {
		if(strcmp(ptr->name, tmp->name) != 0) continue;
		if(ptr->paras == NULL && tmp->paras == NULL) {
			found = true;
			break;
		}
		if(ptr->paras == NULL && tmp->paras != NULL || ptr->paras != NULL && tmp->paras == NULL) 
			continue;
		bool flag = false;
		Variable *ptrPara = ptr->paras->head;
		Variable *tmpPara = tmp->paras->head;
		while(ptrPara && tmpPara) {
			if(strcmp(ptrPara->var, tmpPara->var) != 0) {
				flag = true;
				break;
			}
			ptrPara = ptrPara->next;
			tmpPara = tmpPara->next;
		}
		if(flag) continue;
		if(ptrPara == NULL && tmpPara != NULL || ptrPara != NULL && tmpPara == NULL) 
			continue;
		found = true;
		break;
	}	
	freePredicate(tmp);
	if(!found) return NULL;
	return ptr;
}

MethodStruct* findMethodStruct(MethodStruct *mStruct)
{
	if(methodStructList == NULL) return NULL;
	bool repeat = false;
	MethodStruct *pStruct = methodStructList->head;
	for(; pStruct != NULL; pStruct = pStruct->next) {
		if(!equal(pStruct->method, mStruct->method)) continue;
		if(!equal(pStruct->task, mStruct->task)) continue;
		if(!equal(pStruct->subTasks, mStruct->subTasks)) continue;
		repeat = true;
		break;
	}
	if(!repeat) return NULL;
	return pStruct;
}

MethodStruct* addMethodStruct(MethodStruct *mStruct)
{
	if(methodStructList == NULL) {
		methodStructList = newMethodStructList();
		methodStructList->head = mStruct;
		methodStructList->tail = mStruct;
		return mStruct;	
	} 
	MethodStruct *found = findMethodStruct(mStruct);
	if(found == NULL) {
		methodStructList->tail->next = mStruct;
		methodStructList->tail = mStruct;
		return mStruct;
	}
	freeMethodStruct(mStruct);
	return found;
}

void addLeaf(Action *action)
{	
	if(totalLeafs->tail->head == NULL) {
		totalLeafs->tail->head = action;
		totalLeafs->tail->tail = action;
		return;
	} 
	totalLeafs->tail->tail->glbNext = action;
	action->glbPre = totalLeafs->tail->tail;
	totalLeafs->tail->tail = action;	
}

void addClauseList(ClauseList **list, int &num, Clause *clause, int w, int fw)
{
	if((*list) == NULL) {
		(*list) = newClauseList();
		(*list)->head = clause;
		(*list)->tail = clause;
		(*list)->tail->weight = w;
		(*list)->tail->finalWeight = fw;
		++num;
		return;
	}
	for(Clause *ptr = (*list)->head; ptr != NULL; ptr = ptr->next) {
		if(!equal(ptr->literalList, clause->literalList)) continue;
		ptr->weight += w;
		ptr->finalWeight += fw;
		
		for(Literal *p = clause->literalList->head; p != NULL; p = p->next)
			free(p);
		free(clause->literalList);
		free(clause);
		return;
	}
	(*list)->tail->next = clause;
	(*list)->tail = clause;
	(*list)->tail->weight = w;
	(*list)->tail->finalWeight = fw;
	++num;
}

void addClauseList(ClauseList **list, int &num, Clause *clause, int &weight)
{
	if(weight == 0) makeMaxWeight(*list, weight);
	if((*list) == NULL) {
		(*list) = newClauseList();
		(*list)->head = clause;
		(*list)->tail = clause;
		(*list)->tail->finalWeight = weight;
		++num;
		return;
	}
	
	for(Clause *p = (*list)->head; p != NULL; p = p->next) {
		if(equal(p->literalList, clause->literalList)) {
			p->finalWeight += weight;
			return;
		}	
	}
	
	++num;
	(*list)->tail->next = clause;
	(*list)->tail = clause;
	(*list)->tail->finalWeight = weight;
}

Integer* belong(VariableList *list1, VariableList *list2) 
{
	if(list1 == NULL) {	
		Integer *ans = newInteger();
		ans->len = 0;
		return ans;
	}
	if(list1 != NULL && list2 == NULL) return NULL;
	bool flag = true;
	int args[MAX_VAR];
	int len = 0;
	for(Variable *ptr1 = list1->head; ptr1 != NULL; ptr1 = ptr1->next) {
		bool found = false;
		int count = 1;
		for(Variable *ptr2 = list2->head; ptr2 != NULL; ptr2 = ptr2->next) {
			if(strcmp(ptr1->var, ptr2->var) == 0) {
				found = true;
				args[len] = count;
				++len;
				break;
			}
			++count;
		}
		if(!found) return NULL;
	}
	Integer *ans = newInteger();
	ans->len = len;
	ans->integers = newInt(len);
	int i = 0;
	for(int *p = ans->integers; p != ans->integers+len; ++p) {
		*p = args[i];
		++i;
	}
	return ans;
}

Integer* belong(Integer *list1, int list2[])
{
	if(list1 == NULL) {
		Integer *ans = newInteger();
		ans->len = 0;
		return ans;
	}
	
	bool flag = true;
	int args[MAX_VAR];
	int len = 0;
	for(int i = 0; i != list1->len; ++i) {
		bool found = false;
		for(int j = 1; j < MAX_VAR, list2[j]; ++j) {
			if(*(list1->integers+i) == list2[j]) {
				found = true;
				args[len] = j;
				++len; 
				break;
			}
		}
		if(!found) return NULL;
	}
	Integer *ans = newInteger();
	ans->len = len;
	ans->integers = newInt(len);
	int i = 0;
	for(int *p = ans->integers; p != ans->integers+len; ++p) {
		*p = args[i];
		++i;
	}
	return ans;
}

FormalState* belong(char *name, VariableList *list1, VariableList *list2)
{
	FormalState *ans = newFormalState();
	copystr(&(ans->name), name);
	ans->paras = belong(list1, list2);
	
	return ans;
}

bool simpleBelong(VariableList *list1, VariableList *list2)
{
	if(list1 == NULL) return true;
	if(list1 != NULL && list2 == NULL) return false;
	for(Variable *p1 = list1->head; p1 != NULL; p1 = p1->next) {
		bool found = false;
		for(Variable *p2 = list2->head; p2 != NULL; p2 = p2->next) {
			if(strcmp(p1->var, p2->var) == 0) {
				found = true;
				break;
			}
		}
		if(!found) return false;
	}
	return true;
}

bool simpleBelong(LiteralList *list1, LiteralList *list2)
{
	if(list1 == NULL) return true;
	if(list1 != NULL && list2 == NULL) return false;
	for(Literal *p1 = list1->head; p1 != NULL; p1 = p1->next) {
		bool found = false;
		for(Literal *p2 = list2->head; p2 != NULL; p2 = p2->next) {
			if(p1->isPositive == p2->isPositive && p1->atom->no == p2->atom->no) {
				found = true;
				break;
			}
		}
		if(!found) return false;
	}
	return true;
}

bool equal(Predicate *pred1, Predicate *pred2)
{
	if(strcmp(pred1->name, pred2->name) != 0) return false;
	if(pred1->paras == NULL && pred2->paras == NULL) return true;
	if(pred1->paras != NULL && pred2->paras == NULL) return false;
	if(pred1->paras == NULL && pred2->paras != NULL) return false;
	bool flag = true;
	Variable *ptr1 = pred1->paras->head;
	Variable *ptr2 = pred2->paras->head;
	while(ptr1 && ptr2) {
		if(strcmp(ptr1->var, ptr2->var) != 0) {
			flag = false;
			break;
		}
		ptr1 = ptr1->next;
		ptr2 = ptr2->next;
	}
	if(ptr1 != ptr2) return false;
	return flag;
}

bool equal(TaskStruct *tStruct1, TaskStruct *tStruct2)
{
	if(strcmp(tStruct1->name, tStruct2->name) != 0) return false;
	bool flag = true;
	for(int i = 0; i < MAX_VAR; ++i) {
		if(tStruct1->vars[i] != tStruct2->vars[i]) {
			flag = false;
			break;
		}
	}
	return flag;
}

bool equal(TaskStructList *list1, TaskStructList *list2)
{
	if(list1 == NULL && list2 == NULL) return true;
	if(list1 == NULL && list2 != NULL) return false;
	if(list1 != NULL && list2 == NULL) return false;
	TaskStruct *p1 = list1->head;
	TaskStruct *p2 = list2->head;
	bool flag = true;
	while(p1 && p2) {
		if(!equal(p1, p2)) {
			flag = false;
			break;
		}
		p1 = p1->next;
		p2 = p2->next;
	}
	if(p1 != p2) return false;
	return flag;
}

bool equal(Integer *int1, Integer *int2)
{
	if(int1 == NULL && int2 == NULL) return true;
	if(int1 != NULL && int2 == NULL) {
		if(int1->len != 0) return false;
		else return true;
	}
	if(int1 == NULL && int2 != NULL) {
		if(int2->len != 0) return false;
		else return true;
	}
	
	if(int1->len != int2->len) return false;
	int *p1 = int1->integers;
	int *p2 = int2->integers;
	bool flag = true;
	while(p1 != int1->integers + int1->len) {
		if((*p1) != (*p2)) {
			flag = false;
			break;
		}
		++p1;
		++p2;
	}
	return flag;
}

bool equal(LiteralList *list1, LiteralList *list2)
{
	if(list1 == list2) return true;
	bool belong = true;
	int len1 = 0;
	for(Literal *ptr1 = list1->head; ptr1 != NULL; ptr1 = ptr1->next) {
		bool found = false;
		for(Literal *ptr2 = list2->head; ptr2 != NULL; ptr2 = ptr2->next) {
			if(ptr1->isPositive == ptr2->isPositive && ptr1->atom->no == ptr2->atom->no) {
				found = true;
				break;
			}
		}
		if(!found) {
			belong = false;
			break;
		}
		++len1;
	}
	if(!belong) return false;
	int len2 = 0;
	for(Literal *ptr2 = list2->head; ptr2 != NULL; ptr2 = ptr2->next)
		++len2;
	if(len1 != len2) return false;
	return true;
}

bool equal(FormalState *state1, FormalState *state2)
{
	if(strcmp(state1->name, state2->name) != 0) return false;
	return equal(state1->paras, state2->paras);
}

bool equal(FormalState *state, FormalState *state1, int tran[])
{
	if(strcmp(state->name, state1->name) != 0) return false;
	
	if(state1->paras == NULL) {
		if(state->paras == NULL) return true;
		if(state->paras->len == 0) return true;
		else return false;
	}
	
	Integer *paras = newInteger();
	paras->len = state1->paras->len;
	paras->integers = newInt(paras->len);
	for(int i = 0; i != paras->len; ++i) {
		*(paras->integers+i) = tran[*(state1->paras->integers+i)];
	}
	
	return equal(state->paras, paras);
}

VariableList* getPRS(Action *action, Task *task)
{
	VariableList *ap = action->action->paras;
	VariableList *tp = task->task->paras;
	if(ap == NULL ||  tp == NULL) return NULL;
	VariableList *ans = NULL;
	for(Variable *p1 = ap->head; p1 != NULL; p1 = p1->next) {
		for(Variable *p2 = tp->head; p2 != NULL; p2 = p2->next) {
			if(strcmp(p1->var, p2->var) != 0) continue;
			if(ans == NULL) {
				ans = newVariableList();
				ans->head = newVariable();
				ans->tail = ans->head;
				copystr(&(ans->tail->var), p1->var);
				continue;
			}
			ans->tail->next = newVariable();
			ans->tail = ans->tail->next;
			copystr(&(ans->tail->var), p1->var);
		}
	}
	return ans;
}

Predicate* getFormalPredicate(Predicate *pred)
{
	Predicate *ans = newPredicate();
	copystr(&(ans->name), pred->name);
	
	if(pred->paras == NULL) return ans;
	
	ans->paras = newVariableList();
	ans->paras->head = newVariable();
	ans->paras->tail = ans->paras->head;
	
	Variable *pTail = ans->paras->tail;
	
	for(Variable *pPara = pred->paras->head; pPara != NULL; pPara = pPara->next) {
		bool found = false;
		for(Object *pObj = objectList->head; pObj != NULL; pObj = pObj->next) {
			if(strcmp(pObj->name, pPara->var) == 0) {
				if(pTail->var != NULL) {
					pTail->next = newVariable();
					pTail = pTail->next;
				} 
				copystr(&(pTail->var), pObj->type);
				found = true;
				break;
			}
		}
		if(!found) {
			printf("Error: no find the type of \"%s\" !\n", pPara->var);
			exit(1);
		}
	}	
	return ans;
}

Literal* getLiteral(Atom *atom, bool isPositive)
{
	Literal *ans = newLiteral();
	ans->isPositive = isPositive;
	ans->atom = atom;
	return ans;
}

Clause* getClause(Literal *literal)
{
	Clause *ans = newClause();
	ans->literalList = newLiteralList();
	ans->literalList->head = literal;
	ans->literalList->tail = literal;
	return ans;
}

Clause* getClause(Literal *literal1, Literal *literal2)
{
	Clause *ans = newClause();
	ans->literalList = newLiteralList();
	LiteralList *p = ans->literalList;
	p->head = literal1;
	p->head->next = literal2;
	p->tail = literal2;
	return ans;
}

Integer* getArguments(Predicate *state, Task *task)
{
	if(state->paras == NULL || task->methodParas == NULL) return NULL;
	int len = 0;
	int tmp[MAX_VAR];
	VariableList *sList = state->paras;
	VariableList *tList = task->methodParas;
	for(Variable *ps = sList->head; ps != NULL; ps = ps->next) {
		int count = 1;
		for(Variable *pt = tList->head; pt != NULL; pt = pt->next) {
			if(strcmp(ps->var, pt->var) == 0) {
				tmp[len] = count;
				++len;
				break;
			}
			++count;
		}
	}
	Integer *ans = newInteger();
	ans->len = len;
	ans->integers = newInt(len);
	int i = 0;
	for(int *p = ans->integers; p != ans->integers+len; ++p) {
		*p = tmp[i];
		++i;
	}
	return ans;	
}

PredicateList* getState(PredicateList *list1, PredicateList *list2)
{
	if(list1 == NULL && list2 == NULL) return NULL;
	if(list1 == NULL && list2 != NULL) return NULL;
	if(list1 != NULL && list2 == NULL) {
		if(list1->head == NULL) return NULL;
		
		PredicateList* ans = newPredicateList();
		ans->head = list1->head;
		ans->tail = list1->head;
		for(Predicate *p = list1->head->next; p != NULL; p = p->next) {
			ans->tail->add = p;
			ans->tail = p; 
		}
		return ans;
	}
	PredicateList *ans = NULL;
	for(Predicate *p1 = list1->head; p1 != NULL; p1 = p1->next) {
		bool found = false;
		for(Predicate *p2 = list2->head; p2 != NULL; p2 = p2->next) {
			if(equal(p1, p2)) {
				found = true;
				break;
			}
		}
		if(found) continue;
		if(ans == NULL) {
			ans = newPredicateList();
			ans->head = p1;
			ans->tail = p1;
			continue;
		}
		ans->tail->add = p1;
		ans->tail = p1;
	}
	return ans;	
}

PredicateList* getAddedState(PredicateList *list1, PredicateList *list2)
{
	return getState(list2, list1);
}

PredicateList* getDelState(PredicateList *list1, PredicateList *list2)
{
	return getState(list1, list2);
}

Atom* getActionAtom(Predicate *state, Predicate *action, Integer *args, AtomList **list, int &num, bool obs)
{
	Predicate *tmp1 = findPredicate(state);
	Predicate *tmp2 = findActionSchema(action);
	
	if((*list) == NULL) {
		++num;
		(*list) = newAtomList();
		(*list)->head = newAtom();
		(*list)->tail = (*list)->head;
		(*list)->tail->ptr1 = tmp1;
		(*list)->tail->ptr2 = tmp2;
		(*list)->tail->args = args;	
		(*list)->tail->no = num;
		if(obs) (*list)->tail->obs = obs;
		return (*list)->tail;
	}
	for(Atom *p = (*list)->head; p != NULL; p = p->next) {
		if(p->ptr1 == tmp1 && p->ptr2 == tmp2) {
			if(equal(p->args, args)){
				return p;
			}
		}
	}
	
	++num;
	(*list)->tail->next = newAtom();
	(*list)->tail = (*list)->tail->next;
	(*list)->tail->ptr1 = tmp1;
	(*list)->tail->ptr2 = tmp2;
	(*list)->tail->args = args;
	(*list)->tail->no = num;
	if(obs) (*list)->tail->obs = obs;
	return (*list)->tail;	
}

void buildPath(char name[], char *path, char *file)
{
	if(name == NULL) return;
	strcpy(name, path);
	strcat(name, file);
}

char* intToString(int i) 
{
	char *ans = (char *) calloc(15, sizeof(char));
	int count = 0;
	while(i) {
		ans[count++] = i % 10 + '0';
		i /= 10;
	}
	
	int j = 0;
	int k = count - 1;
	while((j != k) && (j != k-1)) {
		char tmp = ans[k];
		ans[k] = ans[j];
		ans[j] = tmp;
		++j;
		--k;
	}
	if(j == k-1) {
		char tmp = ans[k];
		ans[k] = ans[j];
		ans[j] = tmp;
	}
	
	return ans;
}

int* uniqueRand(int len, int num)
{
	bool flag[len];	
	for(int i = 0 ; i < len; ++i) flag[i] = false;
	
	int count = 0;
	while(count != num) {
		int r = rand() % len;
		if(flag[r] == false) {
			flag[r] = true;
			++count;
		}
	}	
	
	int *ans = (int *) calloc(num, sizeof(int));
	
	int j = 0;
	for(int i = 0; i < len; ++i) {
		if(flag[i]) {
			ans[j] = i;
			++j;
		}
	}
	
	return ans;
}

bool isObsRepeated(Action *action, Predicate *predicate)
{
	if(action == NULL || action->glbNext == NULL) return false;
	if(action->glbNext->observations == NULL) return false;
	
	for(Predicate *p = action->glbNext->observations->head; p != NULL; p = p->next) {
		if(equal(p, predicate)) return true;
	}
	return false;
}

void buildWhiteList()
{
	if(totalLeafs == NULL) return;
	
	for(ActionList *p = totalLeafs->head; p != NULL; p = p->next) {
		if(p->head == NULL) continue;
		Action *h = p->head;
		Predicate *act = h->action;
		for(Predicate *pState = h->observations->head; pState != NULL; pState = pState->next) {
			if(!simpleBelong(pState->paras, act->paras))	continue;																	
			
			bool flag = true;
			FormalState *formalState = belong(pState->name, pState->paras, act->paras);
			
			for(ActionList *p1 = totalLeafs->head; p1 != NULL; p1 = p1->next) {
				if(p == p1 || p1->head == NULL) continue;
				Action *h1 = p1->head;
				Predicate *act1 = h1->action;
				if(strcmp(act->name, act1->name) != 0) continue;
				
				bool found = false;
				for(Predicate *pState1 = h1->observations->head; pState1 != NULL; pState1 = pState1->next) {
					if(!simpleBelong(pState1->paras, act1->paras)) continue;
					
					FormalState *formalState1 = belong(pState1->name, pState1->paras, act1->paras);
					if(equal(formalState, formalState1)) {
						found = true;
						break;
					}					
				}
				if(!found) {
					flag = false;
					break;
				}
			}
			if(!flag) continue;
			
			addActionPred(&whiteList, act->name, formalState);
		}	
	}
}

void addActionPred(ActionPredList **list, char *actionName, FormalState *formalState)
{
	if(*list == NULL) {
		*list = newActionPredList();
		(*list)->head = newActionPred();
		(*list)->tail = (*list)->head;
		
		copystr(&((*list)->tail->actionName), actionName);
		(*list)->tail->stateList = newFormalStateList();
		(*list)->tail->stateList->head = formalState;
		(*list)->tail->stateList->tail = formalState;
		return;
	}
	
	for(ActionPred *p = (*list)->head; p != NULL; p = p->next) {
		if(strcmp(p->actionName, actionName) == 0) {
			for(FormalState *formalState1 = p->stateList->head; formalState1 != NULL; formalState1 = formalState1->next) {
				if(equal(formalState, formalState1)) return;
			}
			
			p->stateList->tail->next = formalState;
			p->stateList->tail = formalState;
			return;
		}
	}	
	
	(*list)->tail->next = newActionPred();
	(*list)->tail = (*list)->tail->next;
	copystr(&((*list)->tail->actionName), actionName);
	(*list)->tail->stateList = newFormalStateList();
	(*list)->tail->stateList->head = formalState;
	(*list)->tail->stateList->tail = formalState;
}

bool isPassWhiteList(Predicate *state, Predicate *action)
{	
	if(!simpleBelong(state->paras, action->paras)) return false;
	if(whiteList == NULL) return true;
	
	bool found = false;	
	for(ActionPred *p = whiteList->head; p != NULL; p = p->next) {
		if(strcmp(p->actionName, action->name) != 0) continue;
		
		found = true;

		FormalState *formalState = belong(state->name, state->paras, action->paras);
		for(FormalState *formalState1 = p->stateList->head; formalState1 != NULL; formalState1= formalState1->next) {
			if(equal(formalState, formalState1)) return true;
		}
	}
	
	if(found) return false;
	return true;
}

bool isPassWhiteList(char *name, Integer *integer, Predicate *action)
{
	if(whiteList == NULL) return true;
	
	bool found = false;	
	for(ActionPred *p = whiteList->head; p != NULL; p = p->next) {
		if(strcmp(p->actionName, action->name) != 0) continue;
		
		found = true;

		FormalState *formalState = newFormalState();
		copystr(&(formalState->name), name);
		formalState->paras = integer;
		for(FormalState *formalState1 = p->stateList->head; formalState1 != NULL; formalState1= formalState1->next) {
			if(equal(formalState, formalState1)) return true;
		}
	}
	
	if(found) return false;
	return true;
}

void buildBlackList()
{
	if(totalLeafs == NULL) return;
	
	Action *guard = newAction();
	PredicateList *goal = goals->head;	
	
	for(ActionList *p = totalLeafs->head; p != NULL; p = p->next) {
		guard->observations = goal;
		p->tail->glbNext = guard;
		
		for(Action *p1 = p->head; p1 != guard; p1 = p1->glbNext) {
			if(p1->observations == NULL) continue;
			for(Predicate *p2 = p1->observations->head; p2 != NULL; p2 = p2->next) {
				if(!simpleBelong(p2->paras, p1->action->paras)) continue;
				if(!isObsRepeated(p1, p2)) continue;

				FormalState *formalState = belong(p2->name, p2->paras, p1->action->paras);
				addActionPred(&blackList, p1->action->name, formalState);
			}
		}
		
		p->tail->glbNext = NULL;
		goal = goal->next;
	}
	free(guard);
}

bool isInBlackList(Predicate *state, Predicate *action)
{
	if(!simpleBelong(state->paras, action->paras)) return true;
	if(blackList == NULL) return false;
	
	for(ActionPred *p = blackList->head; p != NULL; p = p->next) {
		if(strcmp(p->actionName, action->name) != 0) continue;
		
		FormalState *formalState = belong(state->name, state->paras, action->paras);
		for(FormalState *formalState1 = p->stateList->head; formalState1 != NULL; formalState1 = formalState1->next) {
			if(equal(formalState, formalState1)) return true;
		}
	}
	return false;
}

bool isInBlackList(char *name, Integer *integer, Predicate *action)
{
	if(blackList == NULL) return false;
		
	for(ActionPred *p = blackList->head; p != NULL; p = p->next) {
		if(strcmp(p->actionName, action->name) != 0) continue;
		
		FormalState *formalState = newFormalState();
		copystr(&(formalState->name), name);
		formalState->paras = integer;
		for(FormalState *formalState1 = p->stateList->head; formalState1 != NULL; formalState1 = formalState1->next) {
			if(equal(formalState, formalState1)) return true;
		}
	}
	return false;
}

LeafPointer* getLeafPointer(Task *task)
{
	if(task->actions == NULL && task->subTasks == NULL) return NULL;
	
	LeafPointer *ans = newLeafPointer();
	
	if(task->actions == NULL) {
		LeafPointer *tmp1 = getLeafPointer(task->subTasks->head);
		LeafPointer *tmp2 = getLeafPointer(task->subTasks->tail);
		if(tmp1 == NULL || tmp2 == NULL)	return NULL;

		ans->start = tmp1->start;
		ans->end = tmp2->end;
		return ans;
	}
	
	if(task->actions->head->order == 1) {
		ans->start = task->actions->head;
	} else {
		LeafPointer *tmp = getLeafPointer(task->subTasks->head);
		if(tmp == NULL) return NULL;
		ans->start = tmp->start;
	}
	
	if(task->subTasks == NULL) {
		ans->end = task->actions->tail;
		return ans;
	}
	
	if(task->actions->tail->order > task->subTasks->tail->order) {
		ans->end = task->actions->tail;
		return ans;
	}
	
	LeafPointer *tmp = getLeafPointer(task->subTasks->tail);
	if(tmp == NULL) return NULL;
	ans->end = tmp->end;
	return ans; 
}

void buildActionModelList()
{
	if(actionSchemaList == NULL) return;
	actionModelList = newActionModelList();
	
	for(Predicate *p = actionSchemaList->head; p != NULL; p = p->next) {
		if(actionModelList->head == NULL) {
			actionModelList->head = newActionModel();
			actionModelList->tail = actionModelList->head;
		} else {
			actionModelList->tail->next = newActionModel();
			actionModelList->tail = actionModelList->tail->next;
		}
		
		actionModelList->tail->action = p;
		actionModelList->tail->precondList = getFormalStateList(formula2->actionPREList, p);
		actionModelList->tail->addList = getFormalStateList(formula2->actionADDList, p);
		actionModelList->tail->delList = getFormalStateList(formula2->actionDELList, p);
	}
}

FormalStateList* getFormalStateList(AtomList* atomList, Predicate *action)
{
	if(atomList == NULL) return NULL;
	
	FormalStateList *ans = NULL;
	for(Atom *p = atomList->head; p != NULL; p = p->next) {
		if(var_best_value[p->no-1] == FALSE) continue;
		
		Predicate *ptr1 = (Predicate *) p->ptr1;
		Predicate *ptr2 = (Predicate *) p->ptr2;
		if(action != ptr2) continue;
		
		if(ans == NULL) {
			ans = newFormalStateList();
			ans->head = newFormalState();
			ans->tail = ans->head;
		} else {
			ans->tail->next = newFormalState();
			ans->tail = ans->tail->next;
		}
		
		ans->tail->name = ptr1->name;
		ans->tail->paras = p->args;
	}
	
	return ans;
}

void buildMethodModelList()
{
	if(methodStructList == NULL) return;
	for(MethodStruct *p = methodStructList->head; p != NULL; p = p->next) {
		if(p->taskList == NULL) continue;
		
		if(methodModelList == NULL) {
			methodModelList = newMethodModelList();
			methodModelList->head = newMethodModel();
			methodModelList->tail = methodModelList->head;
		} else {
			methodModelList->tail->next = newMethodModel();
			methodModelList->tail = methodModelList->tail->next;
		}
		
		methodModelList->tail->methodStruct = p;
		
		for(Task *t = p->taskList->head; t != NULL; t = t->link) {
			if(t->tried && t->leafPointer == NULL) continue;
			if(t->leafPointer == NULL) {
				t->leafPointer = getLeafPointer(t);
				t->tried = 1;
			}
			if(t->leafPointer == NULL) continue;
			
			PredicateList *obsList = findObservations(t);
			if(obsList == NULL) continue;
			
			for(Predicate *pred = obsList->head; pred != NULL; pred = pred->next) {
				if(!simpleBelong(pred->paras, t->methodParas)) continue;
				
				FormalState *state = belong(pred->name, pred->paras, t->methodParas);
				if(isInFormalList(state, methodModelList->tail->precondList)) continue;
				if(isInFormalList(state, methodModelList->tail->blackList)) continue;
				
				bool found = true;
				for(Task *t1 = p->taskList->head; t1 != NULL; t1 = t1->link) {
					if(t1->tried && t1->leafPointer == NULL) continue;
					if(t1->leafPointer == NULL) {
						t1->leafPointer = getLeafPointer(t1);
						t1->tried = 1;
					}
					if(t1->leafPointer == NULL) continue;
					
					if(!isInActionSubSet(state, t1)) {
						found = false;
						break;
					}
				}			
				
				if(!found) {
					addMethodModelList(&(methodModelList->tail->blackList), state);
				} else {	
					addMethodModelList(&(methodModelList->tail->precondList), state);
				}
			}
		}
	}
	//enhanceMethodModelList();
}

PredicateList* findObservations(Task *task)
{
	if(task->subTasks == NULL) {
		if(task->actions == NULL) return NULL;
		else return task->actions->head->observations;
	}
	
	if(task->actions != NULL) {
		if(task->actions->head->order == 1) {
			return task->actions->head->observations;
		}
	}
	
	return findObservations(task->subTasks->head);
}

bool isInFormalList(FormalState *state, FormalStateList *list)
{
	if(list == NULL) return false;
	
	bool found = false;
	for(FormalState *p = list->head; p != NULL; p = p->next) {
		if(equal(state, p)) {
			found = true;
			break;
		}
	}
	return found;
}

bool isInActionSubSet(FormalState *state, Task *task)
{
	if(actionModelList == NULL) return false;
	
	for(Action *p = task->leafPointer->start; p != task->leafPointer->end->glbNext; p = p->glbNext) {
		if(p->actionModel == NULL) {
			for(ActionModel *a = actionModelList->head; a != NULL; a = a->next) {
				if(strcmp(p->action->name, a->action->name) == 0) {
					p->actionModel = a;
					break;
				}
			}
		}
		
		int tran[MAX_VAR];
		transformParameters(task, p, tran);
		
		if(p->actionModel->precondList != NULL) {
			for(FormalState *f = p->actionModel->precondList->head; f != NULL; f = f->next) {
				if(equal(state, f, tran)) return true;
			}	
		}
		
		if(p->actionModel->addList != NULL) {
			for(FormalState *f = p->actionModel->addList->head; f != NULL; f = f->next) {
				if(equal(state, f, tran)) return false;
			}
		}
	}
	
	return false;	
}

void transformParameters(Task *task, Action *action, int tran[]) 
{
	int i = 1;
	if(action->action->paras == NULL) return;
	
	for(Variable *p = action->action->paras->head; p != NULL; p = p->next) {
		bool found = false;
		int j = 1;
		for(Variable *p1 = task->methodParas->head; p1 != NULL; p1 = p1->next, ++j) {
			if(strcmp(p->var, p1->var) == 0) {
				found = true;
				tran[i] = j;
				++i;
				break;
			}
		}
		if(!found) {
			tran[i] = -1;
			++i;
		}
	}
}

void addMethodModelList(FormalStateList **list, FormalState *state)
{
	if(*list == NULL) {
		*list = newFormalStateList();
		(*list)->head = state;
		(*list)->tail = state;
		return;
	}
	
	for(FormalState *p = (*list)->head; p != NULL; p = p->next) {
		if(equal(p, state)) {
			return;
		}
	}
	
	(*list)->tail->next = state;
	(*list)->tail = (*list)->tail->next;
}

void addLiteral(LiteralList* literalList, Literal *literal) 
{
	if(literalList == NULL) return;
	
	if(literalList->head == NULL) {
		literalList->head = literal;
		literalList->tail = literal;
		return;
	}
	
	for(Literal *p = literalList->head; p != NULL; p = p->next) {
		if(p->isPositive == literal->isPositive && p->atom == literal->atom) return;
	}
	
	literalList->tail->next = literal;
	literalList->tail = literal;
}

PredicateList* getVirtualStates(VariableList *list)
{
	if(list == NULL || objectList == NULL || formalPredList == NULL) return NULL;
	ObjectList *ref = newObjectList();
	for(Variable *p = list->head; p != NULL; p = p->next) {
		for(Object *p1 = objectList->head; p1 != NULL; p1 = p1->next) {
			if(strcmp(p->var, p1->name) == 0) {
				if(ref->head == NULL) {
					ref->head = newObject();
					ref->tail = ref->head;
				} else {
					ref->tail->next = newObject();
					ref->tail = ref->tail->next;
				}
				copystr(&(ref->tail->name), p->var);
				copystr(&(ref->tail->type), p1->type);
				break;
			}
		}
	}
	if(ref->head == NULL) return NULL;
	
	PredicateList *ans = NULL;
	for(Predicate *fp = formalPredList->head; fp != NULL; fp = fp->next) {
		if(fp->paras == NULL) {
			if(ans == NULL) {
				ans = newPredicateList();
				ans->head = newPredicate();
				ans->tail = ans->head;
			} else {
				ans->tail->next = newPredicate();
				ans->tail = ans->tail->next;
			}
			copystr(&(ans->tail->name), fp->name);
			continue;
		}
		
		bool flag = true;
		for(Variable *pVar = fp->paras->head; pVar != NULL; pVar = pVar->next) {
			bool found = false;
			for(Object *pObj = ref->head; pObj != NULL; pObj = pObj->next) {
				if(strcmp(pVar->var, pObj->type) == 0) {
					found = true;
					break;
				}
			}
			if(!found) {
				flag = false;
				break;
			}
		}
		if(!flag) continue;
		
		VariableTable *varTable = newVariableTable();
		buildVirtualVariableTable(varTable, fp->paras, ref);
		
		if(varTable != NULL) {
			for(VariableList *p = varTable->head; p != NULL; p = p->next) {
				if(ans == NULL) {
					ans = newPredicateList();
					ans->head = newPredicate();
					ans->tail = ans->head;
				} else {
					ans->tail->next = newPredicate();
					ans->tail = ans->tail->next;
				}
				ans->tail->name = fp->name;
				ans->tail->paras = p;
			}
		}	
	}

	return ans;
}

void buildVirtualVariableTable(VariableTable *table, VariableList *list, ObjectList *ref)
{
	for(Variable *p = list->head; p != NULL; p = p->next) {
		if(table->head == NULL) {
			for(Object *p1 = ref->head; p1 != NULL; p1 = p1->next) {
				if(strcmp(p->var, p1->type) == 0) {
					if(table->head == NULL) {
						table->head = newVariableList();
						table->tail = table->head;
					} else {
						table->tail->next = newVariableList();
						table->tail = table->tail->next;
					}
					table->tail->head = newVariable();
					table->tail->tail = table->tail->head;
					copystr(&(table->tail->tail->var), p1->name);
				}
			}
			continue;
		}
	
		VariableTable *tmp = newVariableTable();
		for(Object *p1 = ref->head; p1 != NULL; p1 = p1->next) {
			if(strcmp(p->var, p1->type) == 0) {
				for(VariableList *p2 = table->head; p2 != NULL; p2 = p2->next) {
					bool found = false;
					for(Variable *p3 = p2->head; p3 != NULL; p3 = p3->next) {
						if(strcmp(p1->name, p3->var) == 0) {
							found = true;
							break;
						}
					}
					if(found) continue;
					if(tmp->head == NULL) {
						tmp->head = newVariableList();
						tmp->tail = tmp->head;
					} else {
						tmp->tail->next = newVariableList();
						tmp->tail = tmp->tail->next;
					}
					for(Variable *p4 = p2->head; p4 != NULL; p4 = p4->next) {
						if(tmp->tail->head == NULL) {
							tmp->tail->head = newVariable();
							tmp->tail->tail = tmp->tail->head;
						} else {
							tmp->tail->tail->next = newVariable();
							tmp->tail->tail = tmp->tail->tail->next;
						}
						copystr(&(tmp->tail->tail->var), p4->var);
					}
					tmp->tail->tail->next = newVariable();
					tmp->tail->tail = tmp->tail->tail->next;
					copystr(&(tmp->tail->tail->var), p1->name);
				}
			}
		}
		table->head = tmp->head;
		table->tail = tmp->tail;
	}
}

void makeMaxWeight(ClauseList *list, int &weight)
{
	if(list == NULL) return;
	
	for(Clause *p = list->head; p != NULL; p = p->next) {
		if(weight < p->weight) 
			weight = p->weight;
	}
	++weight;
	weight *= WAC;
}

LiteralTable* getCNFClauseList(LiteralTable *table)
{
	if(table == NULL) return NULL;
	LiteralTable *ans = newLiteralTable();
	for(LiteralList *p = table->head; p != NULL; p = p->next) {
		if(p == table->head) {
			for(Literal *p1 = p->head; p1 != NULL; p1 = p1->next) {
				if(ans->head == NULL) {
					ans->head = newLiteralList();
					ans->tail = ans->head;
				} else {
					ans->tail->next = newLiteralList();
					ans->tail = ans->tail->next;
				}
				Literal *tmp = newLiteral();
				tmp->isPositive = p1->isPositive;
				tmp->atom = p1->atom;
				addLiteral(ans->tail, tmp);
			}
			continue;
		} 
		
		LiteralTable *tmp = newLiteralTable();
		for(Literal *p1 = p->head; p1 != NULL; p1 = p1->next) {
			for(LiteralList *p2 = ans->head; p2 != NULL; p2 = p2->next) {
				if(tmp->head == NULL) {
					tmp->head = newLiteralList();
					tmp->tail = tmp->head;
				} else {
					tmp->tail->next = newLiteralList();
					tmp->tail = tmp->tail->next;
				}
				for(Literal *p3 = p2->head; p3 != NULL; p3 = p3->next) {
					Literal *tmpLiteral = newLiteral();
					tmpLiteral->isPositive = p3->isPositive;
					tmpLiteral->atom = p3->atom;
					addLiteral(tmp->tail, tmpLiteral);		
				}
				Literal *tmpLiteral = newLiteral();
				tmpLiteral->isPositive = p1->isPositive;
				tmpLiteral->atom = p1->atom;
				addLiteral(tmp->tail, tmpLiteral);		
			}
		}
		
		ans->head = NULL;
		for(LiteralList *p1 = tmp->head; p1 != NULL; p1 = p1->next) {
			bool flag = true;
			for(LiteralList *p2 = tmp->head; p2 != NULL; p2 = p2->next) {
				if(p1 == p2) continue;
				if(simpleBelong(p2, p1) && !simpleBelong(p1, p2)) {
					flag = false;
					break;
				}
			}
			if(!flag) continue;
			if(ans->head == NULL) {
				ans->head = newLiteralList();
				ans->tail = ans->head;
				ans->tail->head = p1->head;
				ans->tail->tail = p1->tail;
			} else {
				bool found = false;
				for(LiteralList *p3 = ans->head; p3 != NULL; p3 = p3->next) {
					if(equal(p3, p1)) {
						found = true;
						break;
					}
				}
				if(found) continue;
				ans->tail->next = newLiteralList();
				ans->tail = ans->tail->next;
				ans->tail->head = p1->head;
				ans->tail->tail = p1->tail;
			}
		}
	}
	return ans;
}

void buildSecondClauseList()
{
	if(formula1->actionPREList == NULL) return;
	for(Atom *p = formula1->actionPREList->head; p != NULL; p = p->next) {
		if(var_best_value[p->no-1] == FALSE) continue;
		if(p->obs == false) continue;
		Atom *atom = newAtom();
		atom->ptr1 = p->ptr1;
		atom->ptr2 = p->ptr2;
		atom->args = p->args;
		addActionPREList(atom);
		Literal *literal = getLiteral(atom, true);
		Clause *clause = getClause(literal);
		addClauseList(&(formula2->clauseList), formula2->clauseNum, clause, formula2->maxWeight);
	}
}

void addActionPREList(Atom *atom)
{
	if(formula2->actionPREList == NULL) {
		++(formula2->atomNum);
		formula2->actionPREList = newAtomList();
		formula2->actionPREList->head = atom;
		formula2->actionPREList->tail = atom;
		atom->no = formula2->atomNum;
		return;
	}	
	
	for(Atom *p = formula2->actionPREList->head; p != NULL; p = p->next) {
		if(p->ptr1 == atom->ptr1 && p->ptr2 == atom->ptr2) {
			if(p->args == atom->args) return;
		}
	}
	
	++(formula2->atomNum);
	formula2->actionPREList->tail->next = atom;
	formula2->actionPREList->tail = atom;
	atom->no = formula2->atomNum;
}
