/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/


#ifndef _TOOL_
#define _TOOL_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "Datastruct.h"
#include "MemMan.h"
#include "Maxsatz.h"

extern ObjectList *objectList;
extern PredicateList *formalPredList;
extern PredicateList *actualPredList;
extern PredicateList *actionSchemaList;
extern MethodStructList *methodStructList;
extern PredicateTable *goals;
extern TaskList *superRoots;
extern ActionTable *totalLeafs;

extern char *postfix;
extern int postfixLen;

extern bool isFirstObservation;
extern ActionPredList *whiteList;
extern ActionPredList *blackList;

extern my_type var_best_value[tab_variable_size];

extern ActionModelList *actionModelList;
extern MethodModelList *methodModelList;

extern int WSC;
extern int WAC;
extern int WT;

extern Formula *formula1;
extern Formula *formula2;

void copystr(char **des, char *src);

void buildMethodParas(VariableList *ans, VariableList *paras); 

void buildTaskStruct(TaskStruct *task, VariableList *list, VarArray &paras);

void buildTaskOfMethod(MethodStruct *mStruct, VariableList *list, VarArray &paras); 

void buildSubTasksOfMethod(MethodStruct *mStruct, VariableTable *table, VarArray &paras);

Predicate* findPredicate(Predicate *state);

Predicate* findActionSchema(Predicate *action);

Predicate* findPred(Predicate *pred, PredicateList *list);

MethodStruct* findMethodStruct(MethodStruct *mStruct);

MethodStruct* addMethodStruct(MethodStruct *mStruct);

void addLeaf(Action *action);

void addClauseList(ClauseList **list, int &num, Clause *clause, int w, int fw);

void addClauseList(ClauseList **list, int &num, Clause *clause, int &weight);

Integer* belong(VariableList *list1, VariableList *list2); 

Integer* belong(Integer *list1, int list2[]);

FormalState* belong(char *name, VariableList *list1, VariableList *list2);

bool simpleBelong(VariableList *list1, VariableList *list2);

bool simpleBelong(LiteralList *list1, LiteralList *list2);

bool equal(Predicate *pred1, Predicate *pred2);

bool equal(TaskStruct *tStruct1, TaskStruct *tStruct2);

bool equal(TaskStructList *list1, TaskStructList *list2);

bool equal(Integer *int1, Integer *int2);

bool equal(LiteralList *list1, LiteralList *list2);

bool equal(FormalState *state1, FormalState *state2);

bool equal(FormalState *state, FormalState *state1, int tran[]);

VariableList* getPRS(Action *action, Task *task); 

Predicate* getFormalPredicate(Predicate *pred);

Literal* getLiteral(Atom *atom, bool isPositive);

Clause* getClause(Literal *literal);

Clause* getClause(Literal *literal1, Literal *literal2);

Integer* getArguments(Predicate *state, Task *task);

PredicateList* getState(PredicateList *list1, PredicateList *list2);

PredicateList* getAddedState(PredicateList *list1, PredicateList *list2);

PredicateList* getDelState(PredicateList *list1, PredicateList *list2);

Atom* getActionAtom(Predicate *state, Predicate *action, Integer *paras, AtomList **list, int &num, bool obs=false);

Atom* getMethodAtom(Predicate *state, MethodStruct *mStruct, Integer *args);

Atom* getMethodPREAtom(Predicate *state, Task *task);

void buildPath(char *name, char *path, char *file);

char* intToString(int i);

int* uniqueRand(int len, int num);

bool isObsRepeated(Action *action, Predicate *predicate);

void buildWhiteList();

void addActionPred(ActionPredList **list, char *actionName, FormalState *formalState);

bool isPassWhiteList(Predicate *state, Predicate *action);

bool isPassWhiteList(char *name, Integer *integer, Predicate *action);

void buildBlackList();

bool isInBlackList(Predicate *state, Predicate *action);

bool isInBlackList(char *name, Integer *integer, Predicate *action);

LeafPointer* getLeafPointer(Task *task);

void buildActionModelList();

FormalStateList* getFormalStateList(AtomList* atomList, Predicate *action);

void buildMethodModelList();

PredicateList* findObservations(Task *task);

bool isInFormalList(FormalState *state, FormalStateList *list);

bool isInActionSubSet(FormalState *state, Task *task);

void transformParameters(Task *task, Action *action, int tran[]);

void addMethodModelList(FormalStateList **list, FormalState *state);

void addLiteral(LiteralList *literalList, Literal *literal);

PredicateList* getVirtualStates(VariableList *list);

void buildVirtualVariableTable(VariableTable *table, VariableList *list, ObjectList *ref);

void makeMaxWeight(ClauseList *list, int &weight);

LiteralTable* getCNFClauseList(LiteralTable *table);

void buildSecondClauseList();

void addActionPREList(Atom *atom);

#endif
