/************************************************************************
 * Palm lab,
 * Department of Computer Science,
 * Sun Yat-sen University, Guangzhou, China
 * http://www.zsusoft.com/~hankz
 *
 * Copyright 2013, Hankz Hankui Zhuo and Lintao Chen
 *
 *
 * This file is part of HTNML.
 *
 * HTNML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTNML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTNML.  If not, see <http://www.gnu.org/licenses/>.
 *
 ************************************************************************/



#include "Transformation.h"

void buildObjectList(VariableList *list, char *type)
{
	if(list == NULL) return;
	
	if(objectList == NULL) {
		objectList = newObjectList();
		objectList->head = newObject();
		objectList->tail = objectList->head;
	} 
	
	Object *objPtr = objectList->tail;
	
	for(Variable *ptr = list->head; ptr != NULL; ptr = ptr->next) {
		if(objPtr->name != NULL) {
			objPtr->next = newObject();
			objPtr = objPtr->next;
		}
		if(ptr == list->head) {
			objPtr->name = newStr(strlen(ptr->var)+ postfixLen +1);
			strcpy(objPtr->name, ptr->var);
			strcat(objPtr->name, postfix);
		}	 
		else copystr(&(objPtr->name), ptr->var);
		
		copystr(&(objPtr->type), type);
	}
	
	objectList->tail = objPtr;
}

void buildVariableList(VariableList *list, char *var)
{	
	if(list == NULL) return;
	
	if(list->head == NULL) {
		Variable* ptr = newVariable();
		copystr(&(ptr->var), var);
		
		list->head = ptr;
		list->tail = ptr;
		return;
	}
	
	Variable* ptr = newVariable();
	ptr->var = newStr(strlen(var)+ postfixLen +1);
	strcpy(ptr->var, var);
	strcat(ptr->var, postfix);
	
	list->tail->next = ptr;
	list->tail = ptr;
}

void buildTaskVariableList(VariableList *list, char *var)
{
	if(list == NULL) return;
	
	Variable *ptr = newVariable();
	copystr(&(ptr->var), var);
	
	if(list->head == NULL) {
		list->head = ptr;
		list->tail = ptr;
		return;
	}
	
	list->tail->next = ptr;
	list->tail = ptr;
}

void buildFormalPredList(VariableList *varList)
{
	buildPredList(&formalPredList, varList);
}

void buildPredList(PredicateList **list, VariableList *varList)
{
	Predicate *pred = getActualPredicate(varList);
	if(pred == NULL) return;
	
	if(*list == NULL) {
		*list = newPredicateList();
		(*list)->head = getFormalPredicate(pred);
		(*list)->tail = (*list)->head;
	} 
	else {
		Predicate *formalPred = getFormalPredicate(pred);
		
		for(Predicate *ptr = (*list)->head; ptr != NULL; ptr = ptr->next) {
			if(equal(formalPred, ptr)) return;
		}
		
		(*list)->tail->next = formalPred;
		(*list)->tail = (*list)->tail->next;
	}
}

void buildActualPredList(VariableList *varList)
{
	if(actualPredList == NULL) {
		actualPredList = newPredicateList();
		actualPredList->head = getActualPredicate(varList);
		actualPredList->tail = actualPredList->head;
		return;
	}	
	actualPredList->tail->next = getActualPredicate(varList);
	actualPredList->tail = actualPredList->tail->next;
}

void buildVariableTable(VariableTable *table, VariableList *list)
{
	if(table == NULL || list == NULL) return;
	
	if(table->head == NULL) {
		table->head = list;
		table->tail = list;
		return;
	}
	
	table->tail->next = list;
	table->tail = list;
}

void buildDtrList(Task *task)
{
	if(task == NULL) return;
	if(task->actions == NULL && task->subTasks == NULL) return;
	
	if(superRoots->tail->subTasks == NULL) {
		superRoots->tail->subTasks = newTaskList();
		superRoots->tail->subTasks->head = task;
		superRoots->tail->subTasks->tail = task;
		task->preTask = superRoots->tail;
		task->order = 1;
		return;
	}
	
	task->order = superRoots->tail->subTasks->tail->order + 1;
	superRoots->tail->subTasks->tail->next = task;
	superRoots->tail->subTasks->tail = task;
	task->preTask = superRoots->tail;
}

void buildActionSchemaList(VariableList *varList)
{
	buildPredList(&actionSchemaList, varList);
}

void addTopLeaf(VariableTable *table, VariableList *list)
{
	Action *action = getPartAction(table, list);
	action->preTask = superRoots->tail;
	int order = 0;
	if(superRoots->tail->subTasks != NULL) {
		order = superRoots->tail->subTasks->tail->order;
	}
	if(superRoots->tail->actions != NULL) {
		if(order < superRoots->tail->actions->tail->order) {
			order = superRoots->tail->actions->tail->order;
		}
	}
	++order;
	action->order = order;
	
	if(superRoots->tail->actions == NULL) {
		superRoots->tail->actions = newActionList();
		superRoots->tail->actions->head = action;
		superRoots->tail->actions->tail = action;
	} else {
		superRoots->tail->actions->tail->next = action;
		action->pre = superRoots->tail->actions->tail;
		superRoots->tail->actions->tail = action;
	}
	
	addLeaf(action);
}

MethodStruct* addMethodStruct(VariableList *list, VariableTable *table)
{
	if(list == NULL || table == NULL) return NULL;
	
	VarArray paras;
	paras.len = 0;
	
	MethodStruct *mStruct = newMethodStruct();
	mStruct->method = newPredicate();
	copystr(&(mStruct->method->name), list->head->var);
	
	buildTaskOfMethod(mStruct, list, paras);
	buildSubTasksOfMethod(mStruct, table, paras);
	
	mStruct->method->paras = newVariableList();
	mStruct->method->paras->head = newVariable();
	mStruct->method->paras->tail = mStruct->method->paras->head;
	
	for(int i = 0; i < paras.len; ++i) {
		for(Object *pObj = objectList->head; pObj != NULL; pObj = pObj->next) {
			if(strcmp(paras.vars[i], pObj->name) == 0) {
				if(mStruct->method->paras->tail->var == NULL) {
					copystr(&(mStruct->method->paras->tail->var), pObj->type);
				} else {
					mStruct->method->paras->tail->next = newVariable();
					copystr(&(mStruct->method->paras->tail->next->var), pObj->type);
					mStruct->method->paras->tail = mStruct->method->paras->tail->next;
				}
			}
		}
	}
	
	return addMethodStruct(mStruct);
}

VariableTable* getVariableTable(NodeList *list)
{
	if(list == NULL) return NULL;
	
	VariableTable *ans = newVariableTable();
	
	for(Node *ptr = list->head; ptr != NULL; ptr = ptr->next) {
		if(ptr->varList != NULL) {
			if(ans->head == NULL) {
				ans->head = ptr->varList;
				ans->tail = ptr->varList;
			} else {
				ans->tail->next = ptr->varList;
				ans->tail = ptr->varList;
			}
		}
	}
	
	return ans;
}

Task* getPartTask(NodeList *nodeList)
{
	Task *ans = newTask();
	if(nodeList == NULL) return ans;
	
	int count = 0;
	for(Node* ptr = nodeList->head; ptr != NULL; ptr = ptr->next) {
		if(ptr->action != NULL) {
			ptr->action->order = (++count);
			ptr->action->preTask = ans;		
			if(ans->actions == NULL) {
				ans->actions = newActionList();
				ans->actions->head = ptr->action;
				ans->actions->tail = ptr->action;
			} else {
				ans->actions->tail->next = ptr->action;
				ptr->action->pre = ans->actions->tail;
				ans->actions->tail = ptr->action;
			}
		} 
		
		if(ptr->task != NULL) {
			ptr->task->order = (++count);
			ptr->task->preTask = ans;	
			if(ans->subTasks == NULL) {
				ans->subTasks = newTaskList();
				ans->subTasks->head = ptr->task;
				ans->subTasks->tail = ptr->task;
			} else {
				ans->subTasks->tail->next = ptr->task;
				ans->subTasks->tail = ptr->task;
			}
		}
	}
	return ans;
}

Predicate* getActualPredicate(VariableList *list)
{
	if(list == NULL) return NULL;
	Predicate *ans = newPredicate();
	copystr(&(ans->name), list->head->var);
	
	if(list->head->next == NULL) {
		ans->paras = NULL;
		return ans;
	}
	ans->paras = newVariableList();
	ans->paras->head = list->head->next;
	ans->paras->tail = list->tail;
	return ans;
}

VariableList* getMethodParas(VariableList *varList, NodeList *nodeList)
{
	VariableList *ans = NULL;
	if(varList->head != NULL) {
		for(Variable *ptr = varList->head->next; ptr != NULL; ptr = ptr->next) {
			if(ans == NULL) {
				ans = newVariableList();
				ans->head = newVariable();
				ans->tail = ans->head;
			} else {
				ans->tail->next = newVariable();
				ans->tail = ans->tail->next;
			}
			copystr(&(ans->tail->var), ptr->var);
		}
	}
	
	if(nodeList == NULL) return ans;	
	
	for(Node *ptr = nodeList->head; ptr != NULL; ptr = ptr->next) {
		if(ptr->action != NULL) 
			buildMethodParas(ans, ptr->action->action->paras);
		if(ptr->task != NULL) 
			buildMethodParas(ans, ptr->task->task->paras);
	}

	return ans;
}

void buildNodeList(NodeList *list, Node *node)
{
	if(list == NULL) return;
	
	if(list->head == NULL) {
		list->head = node;
		list->tail = node;
		return;
	}
	list->tail->next = node;
	list->tail = node;
}

Action* getPartAction(VariableTable *table, VariableList *list)
{
	Action *ans = newAction();
	ans->action = getActualPredicate(list);
	
	if(table == NULL) return ans;
	
	ans->observations = newPredicateList();
	PredicateList *pObs = ans->observations;
	
	for(VariableList *ptr = table->head; ptr != NULL; ptr = ptr->next) {
		if(pObs->head == NULL) {
			pObs->head = getActualPredicate(ptr);
			pObs->tail = pObs->head;
		} else {
			pObs->tail->next = getActualPredicate(ptr);
			pObs->tail = pObs->tail->next;
		}
	}
	return ans;
}

void buildGoal(VariableTable *table)
{
	if(table == NULL) return;
	
	if(goals == NULL) {
		goals = newPredicateTable();
		goals->head = newPredicateList();
		goals->tail = goals->head;
	} else {
		goals->tail->next = newPredicateList();
		goals->tail = goals->tail->next;
	}
	
	VariableTable *tmp = getPartObservations(table);
	if(tmp == NULL) {
		return;
	}
	
	for(VariableList *list = tmp->head; list != NULL; list = list->next) {
		if(goals->tail->head == NULL) {
			goals->tail->head = getActualPredicate(list);
			goals->tail->tail = goals->tail->head;
			continue;
		}
		goals->tail->tail->next = getActualPredicate(list);
		goals->tail->tail = goals->tail->tail->next;
	}
}

Predicate* getPredicate(char *name, VariableList *list)
{
	Predicate *ans = newPredicate();
	copystr(&(ans->name), name);
	ans->paras = list;
	return ans;
}

bool isStatePositive(char *var)
{
	if(var == NULL ) return false;
	if(strcmp(var, "NOT") == 0) return false;
	return true;
}

VariableTable *getPartObservations(VariableTable *table)
{
	if(table == NULL) return NULL;
	if(isFirstObservation) {
		isFirstObservation = false;
		return table;
	}
	if(partObsPercent == 1.0) return table;
	
	int len = 0;
	for(VariableList *p = table->head; p != NULL; p = p->next) {
		++len;
	}
	int num = int (len * partObsPercent);
	if(num < 1) {
		int r = rand() % 1000;
		if(r < len*partObsPercent*1000) num = 1;
		else return NULL;
	}
	
	int *rand = uniqueRand(len, num);	
	
	int count = 0;
	VariableTable *ans = NULL;
	VariableList *p = table->head;
	for(int i = 0; i < num; ++i) {
		while(rand[i] != count) {
			p = p->next;
			++count;
		}
		
		if(ans == NULL) {
			ans = newVariableTable();
			ans->head = p;
			ans->tail = p;
		} else {
			ans->tail->next = p;
			ans->tail = p;
		}
	}
	ans->tail->next = NULL;
	return ans;
}

bool isComplete()
{
	if(partDtrPercent == 1) return true;
	
	int r = rand() % RAND_NUM;
	if(r < cutPoint) return true;
	else return false;
}

void addTopLeaf(NodeList *list)
{
	if(list == NULL) return;
	
	for(Node *p = list->head; p != NULL; p = p->next) {
		if(p->action == NULL) continue;
		p->action->preTask = superRoots->tail;
	
		int order = 0;
		if(superRoots->tail->subTasks != NULL) {
			order = superRoots->tail->subTasks->tail->order;
		}
		if(superRoots->tail->actions != NULL) {
			if(order < superRoots->tail->actions->tail->order) {
				order = superRoots->tail->actions->tail->order;
			}
		}
		++order;
		p->action->order = order;
	
		if(superRoots->tail->actions == NULL) {	
			superRoots->tail->actions = newActionList();
			superRoots->tail->actions->head = p->action;
			superRoots->tail->actions->tail = p->action;
		} else {
			superRoots->tail->actions->tail->next = p->action;
			p->action->pre = superRoots->tail->actions->tail;
			superRoots->tail->actions->tail = p->action;
		}
	}	
}

void addTopTask(NodeList *list)
{
	if(list == NULL) return;
	
	for(Node *p = list->head; p != NULL; p = p->next) {
		if(p->task == NULL) continue;
		addTopTask(p->task);
	}
}

void addTopTask(Task *task)
{
	if(task->actions == NULL && task->subTasks == NULL)	return;
	
	task->preTask = superRoots->tail;
	
	int order = 0;
	if(superRoots->tail->subTasks != NULL) {
		order = superRoots->tail->subTasks->tail->order;
	}
	if(superRoots->tail->actions != NULL) {
		if(order < superRoots->tail->actions->tail->order) {
			order = superRoots->tail->actions->tail->order;
		}
	}
	++order;
	task->order = order;
	
	if(superRoots->tail->subTasks == NULL) {
		superRoots->tail->subTasks = newTaskList();
		superRoots->tail->subTasks->head = task;
		superRoots->tail->subTasks->tail = task;
	} else {
		superRoots->tail->subTasks->tail->next = task;
		superRoots->tail->subTasks->tail = task;
	}
}

void buildMethodStruct(Task *task, MethodStruct *methodStruct)
{
	if(task == NULL || methodStruct == NULL) return;
	
	if(methodStruct->taskList == NULL) {
		methodStruct->taskList = newTaskList();
		methodStruct->taskList->head = task;
		methodStruct->taskList->tail = task;
		return;		
	}
	
	methodStruct->taskList->tail->link = task;
	methodStruct->taskList->tail = task;
}

